/*
 * Fast USB serial driver
 * by Stephen Blackheath <stephen@blacksapphire.com>
 * Copyright (C) 2006 by Symmetric Systems Ltd.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <asm/uaccess.h>
#include <linux/usb.h>

#ifdef CONFIG_USB_SERIAL_DEBUG
        static int debug = 1;
#else
        static int debug;
#endif

#include "usb-serial.h"


/* ENABLE_FAST_WRITE:
 *   Comment out the #define below to disable fast write functionality.
 */
/*#define ENABLE_FAST_WRITE*/

/*
 * Version Information
 */
#define DRIVER_VERSION "v1.5"
#define DRIVER_AUTHOR "Stephen Blackheath <stephen@blacksapphire.com>"
#if defined(ENABLE_FAST_WRITE)
    #define DRIVER_DESC "Fast USB serial driver (fast read/write)"
#else
    #define DRIVER_DESC "Fast USB serial driver (fast read only)"
#endif

/* function prototypes */
static int  fast_usb_serial_open                (struct usb_serial_port *port, struct file *filp);
static void fast_usb_serial_close               (struct usb_serial_port *port, struct file *filp);
static void fast_usb_serial_read_bulk_callback  (struct urb *urb);
static void fast_usb_serial_write_bulk_callback (struct urb *urb);
static int  fast_usb_serial_write               (struct usb_serial_port *port, int from_user, const unsigned char *buf, int count);
static int  fast_usb_serial_write_room          (struct usb_serial_port *port);
static void fast_usb_serial_shutdown            (struct usb_serial *serial);

static struct usb_device_id id_table [] = {
        { USB_DEVICE(0x413c, 0x8116) },    /* Novatel Wireless Expedite EU730 Modem */
        { USB_DEVICE(0x1410, 0x2400) },    /* Novatel Wireless EU730 HSDPA Modem */
        { USB_DEVICE(0x1410, 0x2410) },    /* Novatel Wireless EU740 HSDPA Modem */
        { USB_DEVICE(0x1410, 0x2420) },    /* Novatel Wireless EU870D HSDPA Modem */
        { }                                /* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, id_table);


static struct usb_serial_device_type fast_usb_serial_device = {
        .owner =                THIS_MODULE,
        .name =                 "Fast USB serial driver",
        .id_table =             id_table,
        .num_interrupt_in =     1,
        .num_bulk_in =          1,
        .num_bulk_out =         2,
        .num_ports =            1,
        .open =                 fast_usb_serial_open,
        .close =                fast_usb_serial_close,
        .write =                fast_usb_serial_write,
        .write_room =           fast_usb_serial_write_room,
        /*
        .read_bulk_callback =   fast_usb_serial_read_bulk_callback,
        .write_bulk_callback =  fast_usb_serial_write_bulk_callback,
        */
        .shutdown =             fast_usb_serial_shutdown,
};

struct fast_usb_serial_urbinfo
{
    struct usb_serial_port* port;
    int urb_index;
};

struct fast_usb_serial_urbentry {
    struct list_head list;  /* Must be first */
    struct usb_serial_port *port;
    unsigned char *bulk_out_buffer;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
    struct work_struct work_write_callback;
#else
    struct tq_struct tqueue_write_callback;
#endif
    struct urb* urb;
};

#define NO_OF_TXQ_URBS 32
#define FAST_USB_SERIAL_MAX_ACTIVE_WRITES 1
#define FAST_USB_SERIAL_MAX_ACTIVE_READS 32
#define MAX_USB_READ_SIZE 2048

struct fast_usb_serial_data
{
#if defined(ENABLE_FAST_WRITE)
    struct list_head txq_pending;
    struct list_head txq_active;
    int txq_active_n;
    struct list_head txq_pool;
    int txq_pool_n;
    int tx_bytes_pending;
    spinlock_t spinlock;        /* locks this structure */
#else
    struct tq_struct write_callback;
    int write_blocked;
#endif
    unsigned char *bulk_in_buffer[FAST_USB_SERIAL_MAX_ACTIVE_READS];
    struct urb* read_urb[FAST_USB_SERIAL_MAX_ACTIVE_READS];
    struct fast_usb_serial_urbinfo read_urbinfo[FAST_USB_SERIAL_MAX_ACTIVE_READS];
};

#if defined(ENABLE_FAST_WRITE)
static void fast_usb_serial_write_callback_softint(void *private)
{
        struct fast_usb_serial_urbentry* ue = (struct fast_usb_serial_urbentry*)private;
        struct usb_serial_port *port = ue->port;
        struct fast_usb_serial_data* fusd = (struct fast_usb_serial_data *)port->private;
        struct tty_struct *tty;

          /* Now that we've finished with this urb, put it back into the pool. */
        {
                unsigned long flags;
                spin_lock_irqsave(&fusd->spinlock, flags);
                list_add_tail(&ue->list, &fusd->txq_pool);
                fusd->txq_pool_n++;
                fusd->tx_bytes_pending -= ue->urb->transfer_buffer_length;
                spin_unlock_irqrestore(&fusd->spinlock, flags);
        }

        tty = port->tty;

          /* If the port has been closed, then do nothing. */
        if (port_paranoia_check (port, __FUNCTION__))
                return;

        if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
                (tty->ldisc.write_wakeup)(tty);

        wake_up_interruptible(&tty->write_wait);
}
#else
static void fast_usb_serial_write_callback_softint(void *private)
{
    struct usb_serial_port *port = (struct usb_serial_port *)private;
    struct fast_usb_serial_data* fusd = (struct fast_usb_serial_data *)port->private;
    struct tty_struct *tty;

    tty = port->tty;

      /* If the port has been closed, then do nothing. */
    if (port_paranoia_check (port, __FUNCTION__)) {
            fusd->write_blocked = 0;
            return;
    }

    spin_lock_irq (&port->spinlock);
    fusd->write_blocked = 0;
    if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
            (tty->ldisc.write_wakeup)(tty);

    wake_up_interruptible(&tty->write_wait);
    spin_unlock_irq (&port->spinlock);
}
#endif

#if defined(ENABLE_FAST_WRITE)
/**
 * Must be called under spin lock.
 */
static void check_write_urb(struct fast_usb_serial_data* fusd)
{
    while (!list_empty(&fusd->txq_pending) && fusd->txq_active_n < FAST_USB_SERIAL_MAX_ACTIVE_WRITES) {
        int result;

        struct fast_usb_serial_urbentry* ue = (struct fast_usb_serial_urbentry*)fusd->txq_pending.next;
        list_del(&ue->list);
        list_add_tail(&ue->list, &fusd->txq_active);
        fusd->txq_active_n++;
        /* send the data out the bulk port */
        result = usb_submit_urb(ue->urb);
        if (result) {
            list_del(&ue->list);  // Detach from txq_active
            list_add_tail(&ue->list, &fusd->txq_pool);
            fusd->txq_active_n--;
            fusd->txq_pool_n++;
            err("%s - failed submitting write urb %p, error %d", __FUNCTION__, ue->urb, result);
            break;
        }
    }
}
#endif

static void fast_usb_serial_write_bulk_callback(struct urb *urb)
{
#if defined(ENABLE_FAST_WRITE)
    struct fast_usb_serial_urbentry* ue = (struct fast_usb_serial_urbentry*)urb->context;
    struct usb_serial_port *port = ue->port;
    struct fast_usb_serial_data* fusd = (struct fast_usb_serial_data *)port->private;
    unsigned long flags;

    spin_lock_irqsave(&fusd->spinlock, flags);
    list_del(&ue->list);  // Remove from txq_active list
    fusd->txq_active_n--;
    check_write_urb(fusd);  // See if we should write anything
    if (urb->status) {
        list_add_tail(&ue->list, &fusd->txq_pool);
        fusd->txq_pool_n++;
        fusd->tx_bytes_pending -= ue->urb->transfer_buffer_length;
    }
    else {
        if (!port_paranoia_check (port, __FUNCTION__))
            schedule_task(&ue->tqueue_write_callback);
    }
    spin_unlock_irqrestore(&fusd->spinlock, flags);
#else
    struct usb_serial_port *port = (struct usb_serial_port *)urb->context;
    struct fast_usb_serial_data* fusd = (struct fast_usb_serial_data *)port->private;

    if (urb->status) {
        fusd->write_blocked = 0;
        printk("urb failed! status was %d\n", urb->status);
    }
    else {
        fusd->write_blocked = 2;
        if (!port_paranoia_check (port, __FUNCTION__)) {
            fusd->write_blocked = 1;
            schedule_task(&fusd->write_callback);
        }
    }
#endif
}

static void fast_usb_serial_read_bulk_callback (struct urb *urb)
{
        struct fast_usb_serial_urbinfo* urbinfo = (struct fast_usb_serial_urbinfo*) urb->context;
        struct usb_serial_port *port = urbinfo->port;
        struct usb_serial *serial = get_usb_serial (port, __FUNCTION__);
        struct tty_struct *tty;
        unsigned char *data = urb->transfer_buffer;
        int i;
        int result;

        dbg("%s - port %d", __FUNCTION__, port->number);

        if (!serial) {
                dbg("%s - bad serial pointer, exiting", __FUNCTION__);
                return;
        }

        if (urb->status) {
                dbg("%s - nonzero read bulk status received: %d", __FUNCTION__, urb->status);
                return;
        }

        //usb_serial_debug_data (__FILE__, __FUNCTION__, urb->actual_length, data);

        tty = port->tty;

        if (tty && urb->actual_length) {
                for (i = 0; i < urb->actual_length ; ++i) {
                        /* if we insert more than TTY_FLIPBUF_SIZE characters, we drop them. */
                        if(tty->flip.count >= TTY_FLIPBUF_SIZE) {
                                tty_flip_buffer_push(tty);
                        }
                        /* this doesn't actually push the data through unless tty->low_latency is set */
                        tty_insert_flip_char(tty, data[i], 0);
                }
                tty_flip_buffer_push(tty);
        }

        /* Continue trying to always read  */
        usb_fill_bulk_urb (urb, serial->dev,
                           usb_rcvbulkpipe (serial->dev,
                                            port->bulk_in_endpointAddress),
                           urb->transfer_buffer,
                           urb->transfer_buffer_length,
                           fast_usb_serial_read_bulk_callback, urbinfo);
        result = usb_submit_urb(urb);
          /* Note:
           * fast_usb_serial.c: fast_usb_serial_read_bulk_callback - failed resubmitting read urb, error -12
           * This error can occur when you close the serial port, and it is not a problem.
           */
        if (result)
                err("%s - failed resubmitting read urb, error %d", __FUNCTION__, result);
}

static int fast_usb_serial_open (struct usb_serial_port *port, struct file *filp)
{
        struct usb_serial       *serial;
        struct fast_usb_serial_data     *fusd;
        int                     result = 0;
        int j;

        if (port_paranoia_check (port, __FUNCTION__))
                return -ENODEV;

        dbg("%s - port %d", __FUNCTION__, port->number);

        serial = get_usb_serial (port, __FUNCTION__);
        if (!serial)
                return -ENODEV;

        fusd = kmalloc( sizeof(struct fast_usb_serial_data), GFP_KERNEL );
        if( !fusd ) {
                err("%s - kmalloc(%Zd) failed.", __FUNCTION__, sizeof(struct fast_usb_serial_data));
                return -ENOMEM;
        }
        memset(fusd, 0, sizeof(struct fast_usb_serial_data));

        port->private = fusd;

#if defined(ENABLE_FAST_WRITE)
        INIT_LIST_HEAD(&fusd->txq_pending);
        INIT_LIST_HEAD(&fusd->txq_active);
        INIT_LIST_HEAD(&fusd->txq_pool);
        for (j = 0; j < NO_OF_TXQ_URBS; j++) {
            struct fast_usb_serial_urbentry* ue = kmalloc(sizeof(struct fast_usb_serial_urbentry), GFP_KERNEL);
            ue->port = port;
            ue->tqueue_write_callback.routine = fast_usb_serial_write_callback_softint;
            ue->tqueue_write_callback.data = ue;
            ue->urb = usb_alloc_urb(0);
            if (!ue->urb) {
                kfree(ue);
                err("%s - No free urbs available", __FUNCTION__);
                return -ENOMEM;
            }
            ue->bulk_out_buffer = kmalloc(port->bulk_out_size, GFP_KERNEL);
            if (!ue->bulk_out_buffer) {
                usb_free_urb(ue->urb);
                kfree(ue);
                err("%s - Couldn't allocate bulk_out_buffer", __FUNCTION__);
                return -ENOMEM;
            }
            usb_fill_bulk_urb(ue->urb, serial->dev, 
                          usb_sndbulkpipe(serial->dev, port->bulk_out_endpointAddress),
                          ue->bulk_out_buffer, port->bulk_out_size,
                          fast_usb_serial_write_bulk_callback, 
                          ue);
            list_add_tail(&ue->list, &fusd->txq_pool);
            fusd->txq_pool_n++;
        }
        spin_lock_init(&fusd->spinlock);
#else
        fusd->write_callback.routine = fast_usb_serial_write_callback_softint;
        fusd->write_callback.data = port;
#endif

        /* force low_latency on so that our tty_push actually forces the data through, 
           otherwise it is scheduled, and with high data rates (like with OHCI) data
           can get lost. */
        if (port->tty)
                port->tty->low_latency = 1;

        /* if we have a bulk interrupt, start reading from it */
        if (serial->num_bulk_in) {
                int buffer_size = MAX_USB_READ_SIZE;
                for (j = 0; j < FAST_USB_SERIAL_MAX_ACTIVE_READS; j++) {
                    fusd->bulk_in_buffer[j] = kmalloc (buffer_size, GFP_KERNEL);
                    if (!fusd->bulk_in_buffer[j]) {
                            err("%s - Couldn't allocate bulk_in_buffer", __FUNCTION__);
                            break;
                    }
                    fusd->read_urb[j] = usb_alloc_urb(0);
                    if (!fusd->read_urb[j]) {
                            err("%s - No free urbs available", __FUNCTION__);
                            break;
                    }
                    fusd->read_urbinfo[j].port = port;
                    fusd->read_urbinfo[j].urb_index = j;
                    usb_fill_bulk_urb(fusd->read_urb[j], serial->dev, 
                                  usb_rcvbulkpipe(serial->dev, port->bulk_in_endpointAddress),
                                  fusd->bulk_in_buffer[j], buffer_size, 
                                  fast_usb_serial_read_bulk_callback,
                                  &fusd->read_urbinfo[j]);
                    result = usb_submit_urb(fusd->read_urb[j]);
                    if (result)
                            err("%s - failed resubmitting read urb, error %d", __FUNCTION__, result);
                }
        }

        return result;
}

#if defined(ENABLE_FAST_WRITE)
static void cancel_active_writes(struct fast_usb_serial_data *fusd)
{
    unsigned long flags;
    spin_lock_irqsave(&fusd->spinlock, flags);
    while (!list_empty(&fusd->txq_active)) {
        spin_unlock_irqrestore(&fusd->spinlock, flags);
          /* Note: Cancelling active writes sometimes does not cause their callbacks to
           * be called immediately.  Since blocking would be required to fix this, I have
           * opted to play it safe and just wait for all active writes to finish, rather
           * than actively cancelling them. */
        mdelay(20);
        spin_lock_irqsave(&fusd->spinlock, flags);
    }
    spin_unlock_irqrestore(&fusd->spinlock, flags);
}
#endif

#if defined(ENABLE_FAST_WRITE)
static void free_txqs(struct fast_usb_serial_data *fusd)
{
    cancel_active_writes(fusd);
      /* Let any softints generated by the USB write callback complete. */
    flush_scheduled_tasks();
    while (!list_empty(&fusd->txq_pending)) {
        struct fast_usb_serial_urbentry* ue = (struct fast_usb_serial_urbentry*) fusd->txq_pending.next;
        list_del(&ue->list);
        usb_free_urb(ue->urb);
        kfree(ue->bulk_out_buffer);
        kfree(ue);
    }
    while (!list_empty(&fusd->txq_pool)) {
        struct fast_usb_serial_urbentry* ue = (struct fast_usb_serial_urbentry*) fusd->txq_pool.next;
        list_del(&ue->list);
        fusd->txq_pool_n--;
        usb_free_urb(ue->urb);
        kfree(ue->bulk_out_buffer);
        kfree(ue);
    }
}
#endif

static void fast_usb_serial_close (struct usb_serial_port *port, struct file * filp)
{
        struct usb_serial       *serial;
        struct fast_usb_serial_data     *fusd;

        if (port_paranoia_check (port, __FUNCTION__))
                return;

        dbg("%s - port %d", __FUNCTION__, port->number);

        serial = get_usb_serial (port, __FUNCTION__);
        if (!serial)
                return;

        fusd = (struct fast_usb_serial_data *)port->private;
        if (fusd) {
#if defined(ENABLE_FAST_WRITE)
                free_txqs(fusd);
#else
                if (serial->num_bulk_out)
                    usb_unlink_urb (port->write_urb);
#endif
                if (serial->num_bulk_in) {
                    int j;
                    for (j = 0; j < FAST_USB_SERIAL_MAX_ACTIVE_READS; j++) {
                        usb_unlink_urb(fusd->read_urb[j]);
                        usb_free_urb(fusd->read_urb[j]);
                    }
                }
                kfree(fusd);
                port->private = NULL;
        }
}

static int fast_usb_serial_write (struct usb_serial_port *port, int from_user, const unsigned char *buf, int count)
{
        struct usb_serial       *serial = port->serial;

#if defined(ENABLE_FAST_WRITE)
        struct fast_usb_serial_data     *fusd   = (struct fast_usb_serial_data   *) port->private;

        int written = 0;
        while (1) {
                unsigned long flags;
                int toCopy = count - written;
                int urb_space, new_length;
                struct fast_usb_serial_urbentry* ue = NULL;
                if (toCopy == 0)
                    break;
                spin_lock_irqsave(&fusd->spinlock, flags);
                  /* If the last queued pending URB has some space left in it, then choose it. */
                if (!list_empty(&fusd->txq_pending)) {
                    struct fast_usb_serial_urbentry* last_ue = (struct fast_usb_serial_urbentry*) fusd->txq_pending.prev;
                    if (last_ue->urb->transfer_buffer_length < port->bulk_out_size) {
                        ue = last_ue;
                          /* Unlink it from the pending list while we fill it up, so it doesn't
                           * get sent until we are finished with it. */
                        list_del(&ue->list);
                        fusd->tx_bytes_pending -= ue->urb->transfer_buffer_length;
                    }
                }
                  /* Otherwise, if we have a free urb in our free pool, pop it off. */
                if (ue == NULL && !list_empty(&fusd->txq_pool)) {
                    ue = (struct fast_usb_serial_urbentry*) fusd->txq_pool.next;
                    list_del(&ue->list);
                    fusd->txq_pool_n--;
                    ue->urb->transfer_buffer_length = 0;  /* Empty it. */
                }
                spin_unlock_irqrestore(&fusd->spinlock, flags);
                if (ue == NULL)
                    break;
                urb_space = port->bulk_out_size - ue->urb->transfer_buffer_length;
                if (toCopy > urb_space)
                    toCopy = urb_space;

                if (from_user) {
                    if (copy_from_user(ue->urb->transfer_buffer+ue->urb->transfer_buffer_length, buf+written, toCopy))
                        return -EFAULT;
                }
                else
                    memcpy (ue->urb->transfer_buffer+ue->urb->transfer_buffer_length, buf+written, toCopy);
                new_length = ue->urb->transfer_buffer_length + toCopy;

                //usb_serial_debug_data(toCopy, ue->urb->transfer_buffer);
                /* set up our urb */
                usb_fill_bulk_urb(ue->urb, serial->dev, 
                              usb_sndbulkpipe(serial->dev, port->bulk_out_endpointAddress),
                              ue->urb->transfer_buffer, new_length,
                              fast_usb_serial_write_bulk_callback, 
                              ue);
                written += toCopy;

                spin_lock_irqsave(&fusd->spinlock, flags);
                list_add_tail(&ue->list, &fusd->txq_pending);
                fusd->tx_bytes_pending += new_length;
                check_write_urb(fusd);
                spin_unlock_irqrestore(&fusd->spinlock, flags);
        }
        return written;
#else
        int result;
        struct fast_usb_serial_data     *fusd   = (struct fast_usb_serial_data   *) port->private;

        dbg("%s - port %d", __FUNCTION__, port->number);

        if (count == 0) {
                dbg("%s - write request of 0 bytes", __FUNCTION__);
                return (0);
        }

        /* only do something if we have a bulk out endpoint */
        if (serial->num_bulk_out && !fusd->write_blocked) {
                if (port->write_urb->status == -EINPROGRESS) {
                        dbg("%s - already writing", __FUNCTION__);
                        return (0);
                }

                count = (count > port->bulk_out_size) ? port->bulk_out_size : count;

                if (from_user) {
                        if (copy_from_user(port->write_urb->transfer_buffer, buf, count))
                                return -EFAULT;
                }
                else {
                        memcpy (port->write_urb->transfer_buffer, buf, count);
                }

                usb_serial_debug_data (__FILE__, __FUNCTION__, count, port->write_urb->transfer_buffer);

                /* set up our urb */
                usb_fill_bulk_urb (port->write_urb, serial->dev,
                                   usb_sndbulkpipe (serial->dev,
                                                    port->bulk_out_endpointAddress),
                                   port->write_urb->transfer_buffer, count,
                                   fast_usb_serial_write_bulk_callback,
                                   port);

                /* send the data out the bulk port */
                fusd->write_blocked = 3;
                result = usb_submit_urb(port->write_urb);
                if (result) {
                        fusd->write_blocked = 0;
                        err("%s - failed submitting write urb, error %d", __FUNCTION__, result);
                }
                else {
                        result = count;
                }

                return result;
        }

        /* no bulk out, so return 0 bytes written */
        return (0);
#endif
}

static int fast_usb_serial_write_room (struct usb_serial_port *port)
{
    struct usb_serial   *serial = port->serial;
#if defined(ENABLE_FAST_WRITE)
    struct fast_usb_serial_data         *fusd   = (struct fast_usb_serial_data   *) port->private;
#endif
    int room = 0;

    {
#if defined(ENABLE_FAST_WRITE)
        unsigned long flags;
        spin_lock_irqsave(&fusd->spinlock, flags);
#endif
        if (!port_paranoia_check (port, __FUNCTION__)) {
            if (serial->num_bulk_out) {
#if defined(ENABLE_FAST_WRITE)
                room = fusd->txq_pool_n * port->bulk_out_size;
#else
                if (port->write_urb->status != -EINPROGRESS)
                        room = port->bulk_out_size;
#endif
            }
        }
        else
            room = -EINVAL;  /* Socket is closed */
#if defined(ENABLE_FAST_WRITE)
        spin_unlock_irqrestore(&fusd->spinlock, flags);
#endif
    }

    return (room);
}

static void fast_usb_serial_shutdown (struct usb_serial *serial)
{
        dbg ("%s", __FUNCTION__);
}


static int __init fast_usb_serial_init (void)
{
        usb_serial_register(&fast_usb_serial_device);
        info(DRIVER_VERSION ":" DRIVER_DESC);
        return 0;
}


static void __exit fast_usb_serial_exit (void)
{
        usb_serial_deregister(&fast_usb_serial_device);
}


module_init(fast_usb_serial_init);
module_exit(fast_usb_serial_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("Proprietary");

MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug enabled or not");
