//=============================================================================
//
//      hal_stub.c
//
//      Helper functions for stub
//
//=============================================================================

//=============================================================================
//
// Purpose:     Helper functions for stub, specific to ROS HAL
// Description: Parts of the GDB stub requirements are provided by
//              the ROS HAL, rather than target and/or board specific
//              code. 
//
//=============================================================================

#include <include/pkgconf/hal.h>
#ifdef RTPKG_RTMON
#include <include/pkgconf/rtmon.h>
#endif

#ifdef RTDBG_HAL_DEBUG_GDB_INCLUDE_STUBS

#include <include/hal/hal_stub.h>           // Our header

#include <include/hal/hal_arch.h>           // HAL_BREAKINST
#include <include/hal/hal_cache.h>          // HAL_xCACHE_x
#include <include/hal/hal_intr.h>           // interrupt disable/restore

#include <include/hal/hal_if.h>             // ROM calling interface
#include <include/hal/hal_misc.h>           // Helper functions

#ifdef RTDBG_HAL_DEBUG_GDB_THREAD_SUPPORT
#include <include/hal/dbg-threads-api.h>    // dbg_currthread_id
#endif

#ifdef USE_LONG_NAMES_FOR_ENUM_REGNAMES
#ifndef PC
#define PC REG_PC
#endif
#ifndef SP
#define SP REG_SP
#endif
#endif

//-----------------------------------------------------------------------------
// Extra ROS data.

// Saved registers.
HAL_SavedRegisters *_hal_registers;
target_register_t registers[NUMREGS];
target_register_t alt_registers[NUMREGS] ;  // Thread or saved process state
target_register_t * _registers = registers; // Pointer to current set of registers
target_register_t orig_registers[NUMREGS];  // Registers to get back to original state

#if defined(HAL_STUB_HW_WATCHPOINT) || defined(HAL_STUB_HW_BREAKPOINT)
static int  _hw_stop_reason;   // Reason we were stopped by hw.

// strings indexed by hw stop reasons defined in hal_stub.h
static const char * const _hw_stop_str[] = {
    "",
    "hbreak",
    "watch",
    "rwatch",
    "awatch"
};

static void *_watch_data_addr; // The data address if stopped by watchpoint
#endif

// Register validity checking
#ifdef RTHWR_REGISTER_VALIDITY_CHECKING
int registers_valid[NUMREGS];
int *_registers_valid = registers_valid;
#endif

#ifndef RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT // this should go away
// Interrupt control.
static volatile __PFI __interruptible_control;
#endif

//-----------------------------------------------------------------------------
// Register access

// Return the currently-saved value corresponding to register REG of
// the exception context.
target_register_t 
get_register (regnames_t reg)
{
    return _registers[reg];
}

#ifdef RTHWR_REGISTER_VALIDITY_CHECKING
// Return the validity of register REG.
int
get_register_valid (regnames_t reg)
{
    return _registers_valid[reg];
}
#endif

// Store VALUE in the register corresponding to WHICH in the exception
// context.
void 
put_register (regnames_t which, target_register_t value)
{
    _registers[which] = value;
}

//-----------------------------------------------------------------------------
// Serial stuff
#ifdef RTPKG_RTMON
extern void ecos_bsp_console_putc(char);
extern char ecos_bsp_console_getc(void);
#endif

#if 0
// Write C to the current serial port.
void 
putDebugChar (int c)
{
#ifdef RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT
    __call_if_debug_procs_t __debug_procs = RTACC_CALL_IF_DEBUG_PROCS();
    RTACC_COMM_IF_PUTC(*__debug_procs, c);
#elif defined(RTPKG_RTMON)
    ecos_bsp_console_putc(c);
#else
    HAL_STUB_PLATFORM_PUT_CHAR(c);
#endif
}

// Read one character from the current serial port.
int 
getDebugChar (void)
{
#ifdef RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT
    __call_if_debug_procs_t __debug_procs = RTACC_CALL_IF_DEBUG_PROCS();
    return RTACC_COMM_IF_GETC(*__debug_procs);
#elif defined(RTPKG_RTMON)
    return ecos_bsp_console_getc();
#else
    return HAL_STUB_PLATFORM_GET_CHAR();
#endif
}

// Flush output channel
void
hal_flush_output(void)
{
#ifdef RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT
    __call_if_debug_procs_t __debug_procs = RTACC_CALL_IF_DEBUG_PROCS();
    RTACC_COMM_IF_CONTROL(*__debug_procs, __COMMCTL_FLUSH_OUTPUT);
#endif
}


// Set the baud rate for the current serial port.
void 
__set_baud_rate (int baud) 
{
#ifdef RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT
    __call_if_debug_procs_t __debug_procs = RTACC_CALL_IF_DEBUG_PROCS();
    RTACC_COMM_IF_CONTROL(*__debug_procs, __COMMCTL_SETBAUD, baud);
#elif defined(RTPKG_RTMON)
    // FIXME!
#else
    HAL_STUB_PLATFORM_SET_BAUD_RATE(baud);
#endif
}

#endif
//-----------------------------------------------------------------------------
// GDB interrupt (BREAK) support.

#ifdef RTDBG_HAL_DEBUG_GDB_BREAK_SUPPORT

#ifndef RTPKG_HAL_ARM

#if (HAL_BREAKINST_SIZE == 1)
typedef rt_uint8 t_inst;
#elif (HAL_BREAKINST_SIZE == 2)
typedef rt_uint16 t_inst;
#elif (HAL_BREAKINST_SIZE == 4)
typedef rt_uint32 t_inst;
#else
#error "Don't know how to handle that size"
#endif

typedef struct
{
  t_inst *targetAddr;
  t_inst savedInstr;
} instrBuffer;

static instrBuffer break_buffer;

volatile int rt_hal_gdb_running_step = 0;

void 
rt_hal_gdb_place_break (target_register_t pc)
{
    rt_hal_gdb_interrupt( pc ); // Let's hope this becomes a drop-through:
}

void 
rt_hal_gdb_interrupt (target_register_t pc)
{
	register RT_ADDRWORD	reg_gp;

    RTARC_HAL_SAVE_GP(reg_gp);

    // Clear flag that we Continued instead of Stepping
    rt_hal_gdb_running_step = 0;
    // and override existing break? So that a ^C takes effect...
    if (NULL != break_buffer.targetAddr)
        rt_hal_gdb_remove_break( (target_register_t)break_buffer.targetAddr );

    if (NULL == break_buffer.targetAddr) {
        break_buffer.targetAddr = (t_inst*) pc;
        break_buffer.savedInstr = *(t_inst*)pc;
        *(t_inst*)pc = (t_inst)HAL_BREAKINST;

        __data_cache(CACHE_FLUSH);
        __instruction_cache(CACHE_FLUSH);
    }

    RTARC_HAL_RESTORE_GP(reg_gp);
}

int 
rt_hal_gdb_remove_break (target_register_t pc)
{
    if ( rt_hal_gdb_running_step )
        return 0;

    if ((t_inst*)pc == break_buffer.targetAddr) {
        *(t_inst*)pc = break_buffer.savedInstr;
        break_buffer.targetAddr = NULL;

        __data_cache(CACHE_FLUSH);
        __instruction_cache(CACHE_FLUSH);
        return 1;
    }
    return 0;
}

int 
rt_hal_gdb_break_is_set (void)
{
    if (NULL != break_buffer.targetAddr) {
        return 1;
    }
    return 0;
}

#endif // RTPKG_HAL_ARM

#endif // RTDBG_HAL_DEBUG_GDB_BREAK_SUPPORT

// Use this function to disable serial interrupts whenever reply
// characters are expected from GDB.  The reason we want to control
// whether the target can be interrupted or not is simply that GDB on
// the host will be sending acknowledge characters/commands while the
// stub is running - if serial interrupts were still active, the
// characters would never reach the (polling) getDebugChar.
static void
interruptible(int state)
{
    static int __interrupts_suspended = 0;

    if (state) {
        __interrupts_suspended--;
        if (0 >= __interrupts_suspended) {
            __interrupts_suspended = 0;
#ifdef RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT // this _check_ should go away
            {
                hal_virtual_comm_table_t* __chan;
                __chan = RTACC_CALL_IF_DEBUG_PROCS();
                RTACC_COMM_IF_CONTROL(*__chan, __COMMCTL_IRQ_ENABLE);
            }
#else                
            if (__interruptible_control)
                __interruptible_control(1);
#endif
        }
    } else {
        __interrupts_suspended++;
        if (1 == __interrupts_suspended)
#ifdef RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT // this _check_ should go away
            {
                hal_virtual_comm_table_t* __chan;
                __chan = RTACC_CALL_IF_DEBUG_PROCS();
                RTACC_COMM_IF_CONTROL(*__chan, __COMMCTL_IRQ_DISABLE);
            }
#else                
            if (__interruptible_control)
                __interruptible_control(0);
#endif
    }
}

//-----------------------------------------------------------------------------
// ROS stub entry and exit magic.

#ifdef RTDBG_HAL_DEBUG_GDB_BREAK_SUPPORT
int rt_hal_gdb_break;
#endif

// Called at stub *kill*
static void 
handle_exception_exit( void )
{
    int i;

    for (i = 0; i < (sizeof(registers)/sizeof(registers[0])); i++)
	registers[i] = orig_registers[i];
}

// Called at stub *entry*
static void 
handle_exception_cleanup( void )
{
    static int orig_registers_set = 0;

    interruptible(0);

    // Expand the HAL_SavedRegisters structure into the GDB register
    // array format.
    HAL_GET_GDB_REGISTERS(&registers[0], _hal_registers);
    _registers = &registers[0];

    if (!orig_registers_set) {
	int i;
	for (i = 0; i < (sizeof(registers)/sizeof(registers[0])); i++)
	    orig_registers[i] = registers[i];
	orig_registers_set = 1;
    }
	
#ifdef HAL_STUB_PLATFORM_STUBS_FIXUP
    // Some architectures may need to fix the PC in case of a partial
    // or fully executed trap instruction. GDB only takes correct action
    // when the PC is pointing to the breakpoint instruction it set.
    // 
    // Most architectures would leave PC pointing at the trap
    // instruction itself though, and so do not need to do anything
    // special.
    HAL_STUB_PLATFORM_STUBS_FIXUP();
#endif

#ifdef RTDBG_HAL_DEBUG_GDB_BREAK_SUPPORT
    // If we continued instead of stepping, when there was a break set
    // ie. we were stepping within a critical region, clear the break, and
    // that flag.  If we stopped for some other reason, this has no effect.
    if ( rt_hal_gdb_running_step ) {
        rt_hal_gdb_running_step = 0;
        rt_hal_gdb_remove_break(get_register (PC));
    }

    // FIXME: (there may be a better way to do this)
    // If we hit a breakpoint set by the gdb interrupt stub, make it
    // seem like an interrupt rather than having hit a breakpoint.
    rt_hal_gdb_break = rt_hal_gdb_remove_break(get_register (PC));
#endif

#if defined(HAL_STUB_HW_WATCHPOINT) || defined(HAL_STUB_HW_BREAKPOINT)
    // For HW watchpoint/breakpoint support, we need to know if we
    // stopped because of watchpoint or hw break. We do that here
    // before GDB has a chance to remove the watchpoints and save
    // the information for later use in building response packets.
    _hw_stop_reason = HAL_STUB_IS_STOPPED_BY_HARDWARE(_watch_data_addr);
#endif    
}

// Called at stub *exit*
static void 
handle_exception_init( void )
{
    // Compact register array again.
    HAL_SET_GDB_REGISTERS(_hal_registers, &registers[0]);

    interruptible(1);
}


//-----------------------------------------------------------------------------
// Initialization.

// Signal handler.
int 
rt_hal_process_signal (int signal)
{
    // We don't care about the signal (atm).
    return 0;
}

// Install the standard set of trap handlers for the stub.
void 
__install_traps (void)
{
    // Set signal handling vector so we can treat 'C<signum>' as 'c'.
    __process_signal_vec = &rt_hal_process_signal;
    // Set exit vector to reset vector. This will allow automatic reset on
    // some platforms.
    __process_exit_vec = &handle_exception_exit;

    __cleanup_vec = &handle_exception_cleanup;
    __init_vec    = &handle_exception_init;

#ifndef RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT // this should go away
#ifdef RTDBG_HAL_DEBUG_GDB_BREAK_SUPPORT
    // Control of GDB interrupts.
    __interruptible_control = HAL_STUB_PLATFORM_INTERRUPTIBLE;
#endif
#endif

    // Nothing further to do, handle_exception will be called when an
    // exception occurs.
}

#if 0
// Initialize the hardware.
void 
initHardware (void) 
{
    static int initialized = 0;

    if (initialized)
        return;
    initialized = 1;

    // Get serial port initialized.
    HAL_STUB_PLATFORM_INIT_SERIAL();

#ifdef HAL_STUB_PLATFORM_INIT
    // If the platform defines any initialization code, call it here.
    HAL_STUB_PLATFORM_INIT();
#endif        

#ifndef RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT // this should go away
#ifdef RTDBG_HAL_DEBUG_GDB_BREAK_SUPPORT
    // Get interrupt handler initialized.
    HAL_STUB_PLATFORM_INIT_BREAK_IRQ();
#endif
#endif // !RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT
}
#endif


// Reset the board.
void 
__reset (void)
{
#ifdef RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT
    __call_if_reset_t __rom_reset = RTACC_CALL_IF_RESET_GET();
    if (__rom_reset)
        __rom_reset();
#else
//    HAL_PLATFORM_RESET();
	SmartReboot();
#endif
}

//-----------------------------------------------------------------------------
// Breakpoint support.

#ifndef RTPKG_HAL_ARM
// This function will generate a breakpoint exception.  It is used at
// the beginning of a program to sync up with a debugger and can be
// used otherwise as a quick means to stop program execution and
// "break" into the debugger.
void
breakpoint()
{
    //HAL_BREAKPOINT(_breakinst);
    //__asm("break 5");
    asm  (" .globl  " "_breakinst" ";"          \
             "_breakinst" ":"                       \
              " break   5"                      \
    );
}

// This function returns the opcode for a 'trap' instruction.
unsigned long 
__break_opcode ()
{
  return HAL_BREAKINST;
}
#endif


//-----------------------------------------------------------------------------
// Write the 'T' packet in BUFFER. SIGVAL is the signal the program received.
void 
__build_t_packet (int sigval, char *buf)
{
    target_register_t addr;
    char *ptr = buf;
    target_register_t extend_val = 0;

    *ptr++ = 'T';
    *ptr++ = __tohex (sigval >> 4);
    *ptr++ = __tohex (sigval);

#ifdef RTDBG_HAL_DEBUG_GDB_THREAD_SUPPORT
    // Include thread ID if thread manipulation is required.
    {
        int id;
        
        *ptr++ = 't';
        *ptr++ = 'h';
        *ptr++ = 'r';
        *ptr++ = 'e';
        *ptr++ = 'a';
        *ptr++ = 'd';
        *ptr++ = ':';

#if (RT_BYTEORDER == RT_MSBFIRST)
        id = dbg_currthread_id ();
#else
        // FIXME: Temporary workaround for PR 18903. Thread ID must be
        // big-endian in the T packet.
        {
            unsigned char* bep = (unsigned char*)&id;
            int be_id;

            be_id = dbg_currthread_id ();
            *bep++ = (be_id >> 24) & 0xff ;
            *bep++ = (be_id >> 16) & 0xff ;
            *bep++ = (be_id >> 8) & 0xff ;
            *bep++ = (be_id & 0xff) ;
        }
#endif
        ptr = __mem2hex((char *)&id, ptr, sizeof(id), 0);
        *ptr++ = ';';
    }
#endif

#ifdef HAL_STUB_HW_WATCHPOINT
    switch(_hw_stop_reason) {
      case HAL_STUB_HW_STOP_WATCH:
      case HAL_STUB_HW_STOP_RWATCH:
      case HAL_STUB_HW_STOP_AWATCH:
	strcpy(ptr, _hw_stop_str[_hw_stop_reason]);
	ptr += strlen(_hw_stop_str[_hw_stop_reason]);
	*ptr++ = ':';
	// Send address MSB first
	ptr += __intToHex(ptr, (target_register_t)_watch_data_addr,
			  sizeof(_watch_data_addr) * 8);
	*ptr++ = ';';
	break;
      default:
	break;
    }
#endif

    *ptr++ = __tohex (PC >> 4);
    *ptr++ = __tohex (PC);
    *ptr++ = ':';
    addr = get_register (PC);
    if (sizeof(addr) < REGSIZE(PC))
    {
        // GDB is expecting REGSIZE(PC) number of bytes.
        // We only have sizeof(addr) number.  Let's fill
        // the appropriate number of bytes intelligently.
#ifdef RTARC_SIGN_EXTEND_REGISTERS
        {
            unsigned long bits_in_addr = (sizeof(addr) << 3);  // ie Size in bytes * 8
            target_register_t sign_bit_mask = (1 << (bits_in_addr - 1));
            if ((addr & sign_bit_mask) == sign_bit_mask)
                extend_val = ~0;
        }
#endif
    }
#if (RT_BYTEORDER == RT_MSBFIRST)
    ptr = __mem2hex((char *)&extend_val, ptr, REGSIZE(PC) - sizeof(addr), 0);
#endif
    ptr = __mem2hex((char *)&addr, ptr, sizeof(addr), 0);
#if (RT_BYTEORDER == RT_LSBFIRST)
    ptr = __mem2hex((char *)&extend_val, ptr, REGSIZE(PC) - sizeof(addr), 0);
#endif
    *ptr++ = ';';

    *ptr++ = __tohex (SP >> 4);
    *ptr++ = __tohex (SP);
    *ptr++ = ':';
    addr = (target_register_t) get_register (SP);
    if (sizeof(addr) < REGSIZE(SP))
    {
        // GDB is expecting REGSIZE(SP) number of bytes.
        // We only have sizeof(addr) number.  Let's fill
        // the appropriate number of bytes intelligently.
        target_register_t extend_val = 0;
#ifdef RTARC_SIGN_EXTEND_REGISTERS
        {
            unsigned long bits_in_addr = (sizeof(addr) << 3);  // ie Size in bytes * 8
            target_register_t sign_bit_mask = (1 << (bits_in_addr - 1));
            if ((addr & sign_bit_mask) == sign_bit_mask)
                extend_val = ~0;
        }
#endif
        ptr = __mem2hex((char *)&extend_val, ptr, REGSIZE(SP) - sizeof(addr), 0);
    }
    ptr = __mem2hex((char *)&addr, ptr, sizeof(addr), 0);
    *ptr++ = ';';
    
    *ptr++ = 0;
}

//-----------------------------------------------------------------------------
// Cache functions.

// Perform the specified operation on the instruction cache. 
// Returns 1 if the cache is enabled, 0 otherwise.
int 
__instruction_cache (cache_control_t request)
{
    int state = 1;

    switch (request) {
    case CACHE_ENABLE:
        HAL_ICACHE_ENABLE();
        break;
    case CACHE_DISABLE:
        HAL_ICACHE_DISABLE();
        state = 0;
        break;
    case CACHE_FLUSH:
        HAL_ICACHE_SYNC();
        break;
    case CACHE_NOOP:
        /* fall through */
    default:
        break;
    }

#ifdef HAL_ICACHE_IS_ENABLED
    HAL_ICACHE_IS_ENABLED(state);
#endif

    return state;
}

// Perform the specified operation on the data cache. 
// Returns 1 if the cache is enabled, 0 otherwise.
int 
__data_cache (cache_control_t request)
{
    int state = 1;

    switch (request) {
    case CACHE_ENABLE:
        HAL_DCACHE_ENABLE();
        break;
    case CACHE_DISABLE:
        HAL_DCACHE_DISABLE();
        state = 0;
        break;
    case CACHE_FLUSH:
        HAL_DCACHE_SYNC();
        break;
    case CACHE_NOOP:
        /* fall through */
    default:
        break;
    }
#ifdef HAL_DCACHE_IS_ENABLED
    HAL_DCACHE_IS_ENABLED(state);
#endif

    return state;
}

//-----------------------------------------------------------------------------
// Memory accessor functions.

// The __mem_fault_handler pointer is volatile since it is only
// set/cleared by the function below - which does not rely on any
// other functions, so the compiler may decide to not bother updating
// the pointer at all. If any of the memory accesses cause an
// exception, the pointer must be set to ensure the exception handler
// can make use of it.

void* volatile __mem_fault_handler = (void *)0;

/* These are the "arguments" to __do_read_mem and __do_write_mem, 
   which are passed as globals to avoid squeezing them thru
   __set_mem_fault_trap.  */





//#define ENTR_OF_APPL_PROG	0x80100000	// the default entrant of appl prog
static int flash_erase_flag=0;           //the flag for if the sector of the first write addr in flash is erased,should erase first!    
static int zip_file_flag=0;              //the flag for the current appl program is zip format or not
//these two flag used in __do_write_rom


static volatile target_register_t memCount;

static void
__do_copy_mem (unsigned char* src, unsigned char* dst)
{
    unsigned long *long_dst;
    unsigned long *long_src;
    unsigned short *short_dst;
    unsigned short *short_src;

    __mem_fault = 1;                      /* Defaults to 'fail'. Is cleared */
                                          /* when the copy loop completes.  */
//    __mem_fault_handler = &&err;

    // See if it's safe to do multi-byte, aligned operations
    while (memCount) {
        if ((memCount >= sizeof(long)) &&
            (((target_register_t)dst & (sizeof(long)-1)) == 0) &&
            (((target_register_t)src & (sizeof(long)-1)) == 0)) {
        
            long_dst = (unsigned long *)dst;
            long_src = (unsigned long *)src;

            *long_dst++ = *long_src++;
            memCount -= sizeof(long);

            dst = (unsigned char *)long_dst;
            src = (unsigned char *)long_src;
        } else if ((memCount >= sizeof(short)) &&
                   (((target_register_t)dst & (sizeof(short)-1)) == 0) &&
                   (((target_register_t)src & (sizeof(short)-1)) == 0)) {
            
            short_dst = (unsigned short *)dst;
            short_src = (unsigned short *)src;

            *short_dst++ = *short_src++;
            memCount -= sizeof(short);

            dst = (unsigned char *)short_dst;
            src = (unsigned char *)short_src;
        } else {
            *dst++ = *src++;
            memCount--;
        }
    }

    __mem_fault = 0;

 err:
    __mem_fault_handler = (void *)0;
}

#if 0
//add by shao
extern unsigned int flashaddrstart; 
extern unsigned int flashaddrend; 
static unsigned char *dirtyaddrstart;         //the addr of where the flash has been writed
static unsigned char *dirtyaddrend;          //used to juge if this area need to erase

static void __do_write_rom(unsigned char* src, unsigned char* dst)
{
    if (dirtyaddrstart<(unsigned char*)flashaddrstart||dirtyaddrend>(unsigned char*)flashaddrend)              //initial the dirty area
    {
	dirtyaddrstart=dst;
	dirtyaddrend=dst+memCount;
	flash_erase_sector(dst);
    }
//    else if((dirtyaddrstart<=dst<dirtyaddrend)||(dirtyaddrstart<=dst+memCount<dirtyaddrend)||	// modified by zhao
	else if(((dirtyaddrstart<=dst)&&(dst<dirtyaddrend))||((dirtyaddrstart<=dst+memCount)&&(dst+memCount<dirtyaddrend))||\
	    (dst<dirtyaddrstart&&(get_sector_base(dst)!=get_sector_base(dirtyaddrstart)))||\
	    (dst>dirtyaddrend&&(get_sector_base(dst)!=get_sector_base(dirtyaddrend))))
    {
	
	flash_erase_sector(dst);                                               //judge if need to erase
	dirtyaddrstart=dst;
	dirtyaddrend=dst+memCount;
	
    }
    if(dst<dirtyaddrstart)
    {
  	dirtyaddrstart=dst;
    }
    if(dst+memCount>dirtyaddrend)
	dirtyaddrend=dst+memCount;

    if( get_sector_base(dst) != get_sector_base( dst + memCount) )   //modified by shao
    {
	flash_erase_sector(dst+memCount);
	flash_erase_flag=0;
	//if the memDst has been in the next sector, the first sector should been open for erase for next download
		
    }

/*    if(memCount%2)
    {
	src[memCount]=0xFF;
	memCount++;
    }
    */    
   flash_write(dst,src,memCount);   //should add some flag for faild write flash in future by shao
 
}
#endif

/*
 * __read_mem_safe:
 * Get contents of target memory, abort on error.
 */

int
__read_mem_safe (void *dst, void *src, int count)
{
  memCount = count;
  __do_copy_mem((unsigned char*) src, (unsigned char*) dst);
  return count - memCount;      // return number of bytes successfully read
}

/*
 * __write_mem_safe:
 * Set contents of target memory, abort on error.
 */

int
__write_mem_safe (void *src, void *dst, int count)
{
 
  memCount = count;

  if (memCount>0)
  {
/* 
      if(dst>=(void*)flashaddrstart&&dst<(void*)flashaddrend)
      {
        __do_write_rom((unsigned char*) src, (unsigned char*) dst);
		memCount=0;
      }
      else
	__do_copy_mem((unsigned char*) src, (unsigned char*) dst);
*/
	__do_copy_mem((unsigned char*) src, (unsigned char*) dst);
  }

  return count - memCount;      // return number of bytes successfully written
}

//-----------------------------------------------------------------------------
// Target extras?!
int 
__process_target_query(char * pkt, char * out, int maxOut)
{ return 0 ; }
int 
__process_target_set(char * pkt, char * out, int maxout)
{ return 0 ; }
int 
__process_target_packet(char * pkt, char * out, int maxout)
{ return 0 ; }

// GDB string output, making sure interrupts are disabled.
// This function gets used by some diag output functions.
void 
hal_output_gdb_string(target_register_t str, int string_len)
{
    register unsigned long __state;
    HAL_DISABLE_INTERRUPTS(__state);
/*
	asm volatile ("move $9,$31");
	asm volatile ("jalx ____hal_disable_interrupts;nop");
____hal_disable_interrupts_return:
	asm volatile ("move $31,$9"	\
		"move %0,$8;"	\
		"and    %0,%0,0X1;"	\
		: "=r"(__state)	\
		: "r"(__state)	\
		);
*/
    __output_gdb_string(str, string_len);
    HAL_RESTORE_INTERRUPTS(__state);
/*
	asm volatile(
		"and	%0,%0,0x1;"	\
		"move $8,%0;"	\
		: "=r"(__state)	\
		: "r"(__state)	\
		);
	asm volatile("move $9,$31;jalx ____hal_restore_interrupts;nop");
____hal_restore_interrupts_return:
	asm volatile ("move $31,$9");
*/
}

#endif // RTDBG_HAL_DEBUG_GDB_INCLUDE_STUBS
