#ifndef RTONCE_HAL_HAL_IF_H
#define RTONCE_HAL_HAL_IF_H

//=============================================================================
//
//      hal_if.h
//
//      HAL header for ROM/RAM calling interface.
//
//=============================================================================

//=============================================================================
//
// Purpose:     HAL RAM/ROM calling interface
// Description: ROM/RAM calling interface table	definitions. The layout is a
//              combination of libbsp and vectors already in use by some
//              ROS platforms.
// Usage:       #include <rt/hal/hal_if.h>
//
//=============================================================================

#include <include/infra/rt_type.h>         // types & externC
#include <include/hal/dbg-threads-api.h>
#include <include/hal/dbg-thread-syscall.h>

#include <stdarg.h>

#ifdef RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT

// Architecture/var/platform may override the accessor macros.
#include <rt/hal/hal_arch.h>

//--------------------------------------------------------------------------
#ifndef _BSP_HANDLER_T_DEFINED
#define _BSP_HANDLER_T_DEFINED
typedef int (*bsp_handler_t)(int __irq_nr, void *__regs);
#endif // _BSP_HANDLER_T_DEFINED

//--------------------------------------------------------------------------
// Communication interface table. RTNUM_CALL_IF_CONSOLE_PROCS and
// RTNUM_CALL_IF_DEBUG_PROCS point to instances (possibly the same)
// of this table.

typedef enum {
    /*
     * For serial ports, the control function may be used to set and get the
     * current baud rate. Usage:
     * 
     *   err = (*__control)(COMMCTL_SETBAUD, int bits_per_second);
     *     err => Zero if successful, -1 if error.
     *
     *   baud = (*__control)(COMMCTL_GETBAUD);
     *     baud => -1 if error, current baud otherwise.
     */
    __COMMCTL_SETBAUD=0,
    __COMMCTL_GETBAUD,

    /*
     * Install and remove debugger interrupt handlers. These are the receiver
     * interrupt routines which are used to change control from a running
     * program to the debugger stub.
     */
    __COMMCTL_INSTALL_DBG_ISR,
    __COMMCTL_REMOVE_DBG_ISR,

    /*
     * Disable comm port interrupt. Returns TRUE if interrupt was enabled,
     * FALSE otherwise.
     */
    __COMMCTL_IRQ_DISABLE,

    /*
     * Enable comm port interrupt.
     */
    __COMMCTL_IRQ_ENABLE,

    /*
     * Returns the number of the interrupt vector used by the debug
     * interrupt handler.
     */
    __COMMCTL_DBG_ISR_VECTOR,

    /*
     * Returns the current timeout value and sets a new timeout.
     * Timeout resolution is in milliseconds.
     *   old_timeout = (*__control)(__COMMCTL_SET_TIMEOUT, 
     *                              rt_int32 new_timeout);
     */
    __COMMCTL_SET_TIMEOUT,

    /*
     * Forces driver to send all characters which may be buffered in
     * the driver. This only flushes the driver buffers, not necessarily
     * any hardware FIFO, etc.
     */
    __COMMCTL_FLUSH_OUTPUT,

} __comm_control_cmd_t;


#define RTNUM_COMM_IF_CH_DATA                    0
#define RTNUM_COMM_IF_WRITE                      1
#define RTNUM_COMM_IF_READ                       2
#define RTNUM_COMM_IF_PUTC                       3
#define RTNUM_COMM_IF_GETC                       4
#define RTNUM_COMM_IF_CONTROL                    5
#define RTNUM_COMM_IF_DBG_ISR                    6
#define RTNUM_COMM_IF_GETC_TIMEOUT               7

#define RTNUM_COMM_IF_TABLE_SIZE                 8

typedef volatile RT_ADDRWORD hal_virtual_comm_table_t[RTNUM_COMM_IF_TABLE_SIZE];

// The below is a (messy) attempt at adding some type safety to the
// above array. At the same time, the accessors allow the
// implementation to be easily changed in the future (both tag->data
// table and structure implementations have been suggested).

typedef void* __comm_if_ch_data_t;
typedef void (*__comm_if_write_t)(void* __ch_data, const rt_uint8* __buf,
                                  rt_uint32 __len);
typedef int (*__comm_if_read_t)(void* __ch_data, rt_uint8* __buf,
                                rt_uint32 __len);
typedef void (*__comm_if_putc_t)(void* __ch_data, rt_uint8 __ch);
typedef rt_uint8 (*__comm_if_getc_t)(void* __ch_data);
typedef int (*__comm_if_control_t)(void *__ch_data, 
                                   __comm_control_cmd_t __func, ...);
typedef int (*__comm_if_dbg_isr_t)(void *__ch_data, 
                               int* __ctrlc, RT_ADDRWORD __vector,
                               RT_ADDRWORD __data);
typedef rt_bool (*__comm_if_getc_timeout_t)(void* __ch_data, rt_uint8* __ch);

#ifndef RTACC_COMM_IF_DEFINED
#define RTACC_COMM_IF_CH_DATA(_t_) \
 ((__comm_if_ch_data_t)((_t_)[RTNUM_COMM_IF_CH_DATA]))
#define RTACC_COMM_IF_CH_DATA_SET(_t_, _x_) \
 (_t_)[RTNUM_COMM_IF_CH_DATA]=(RT_ADDRWORD)(_x_)

#define RTACC_COMM_IF_WRITE(_t_, _b_, _l_) \
 ((__comm_if_write_t)((_t_)[RTNUM_COMM_IF_WRITE]))(RTACC_COMM_IF_CH_DATA(_t_), (_b_), (_l_))
#define RTACC_COMM_IF_WRITE_SET(_t_, _x_) \
 (_t_)[RTNUM_COMM_IF_WRITE]=(RT_ADDRWORD)(_x_)

#define RTACC_COMM_IF_READ(_t_, _b_, _l_) \
 ((__comm_if_read_t)((_t_)[RTNUM_COMM_IF_READ]))(RTACC_COMM_IF_CH_DATA(_t_), (_b_), (_l_))
#define RTACC_COMM_IF_READ_SET(_t_, _x_) \
 (_t_)[RTNUM_COMM_IF_READ]=(RT_ADDRWORD)(_x_)

#define RTACC_COMM_IF_PUTC(_t_, _c_) \
 ((__comm_if_putc_t)((_t_)[RTNUM_COMM_IF_PUTC]))(RTACC_COMM_IF_CH_DATA(_t_), (_c_))
#define RTACC_COMM_IF_PUTC_SET(_t_, _x_) \
 (_t_)[RTNUM_COMM_IF_PUTC]=(RT_ADDRWORD)(_x_)

#define RTACC_COMM_IF_GETC(_t_) \
 ((__comm_if_getc_t)((_t_)[RTNUM_COMM_IF_GETC]))(RTACC_COMM_IF_CH_DATA(_t_))
#define RTACC_COMM_IF_GETC_SET(_t_, _x_) \
 (_t_)[RTNUM_COMM_IF_GETC]=(RT_ADDRWORD)(_x_)

#define RTACC_COMM_IF_CONTROL(_t_, args...) \
 ((__comm_if_control_t)((_t_)[RTNUM_COMM_IF_CONTROL]))(RTACC_COMM_IF_CH_DATA(_t_), args)
#define RTACC_COMM_IF_CONTROL_SET(_t_, _x_) \
 (_t_)[RTNUM_COMM_IF_CONTROL]=(RT_ADDRWORD)(_x_)

#define RTACC_COMM_IF_DBG_ISR(_t_, _c_, _v_, _d_) \
 ((__comm_if_dbg_isr_t)((_t_)[RTNUM_COMM_IF_DBG_ISR]))(RTACC_COMM_IF_CH_DATA(_t_), (_c_), (_v_), (_d_))
#define RTACC_COMM_IF_DBG_ISR_SET(_t_, _x_) \
 (_t_)[RTNUM_COMM_IF_DBG_ISR]=(RT_ADDRWORD)(_x_)

#define RTACC_COMM_IF_GETC_TIMEOUT(_t_, _c_) \
 ((__comm_if_getc_timeout_t)((_t_)[RTNUM_COMM_IF_GETC_TIMEOUT]))(RTACC_COMM_IF_CH_DATA(_t_), (_c_))
#define RTACC_COMM_IF_GETC_TIMEOUT_SET(_t_, _x_) \
 (_t_)[RTNUM_COMM_IF_GETC_TIMEOUT]=(RT_ADDRWORD)(_x_)

#endif // RTACC_COMM_IF_DEFINED

//--------------------------------------------------------------------------
// Main calling interface table. Will be assigned a location by the 
// linker script. Both ROM and RAM startup applications will know about
// the location.
#define RTNUM_CALL_IF_VERSION                    0
#define RTNUM_CALL_IF_ICTRL_TABLE                1
#define RTNUM_CALL_IF_EXC_TABLE                  2
#define RTNUM_CALL_IF_DBG_VECTOR                 3
#define RTNUM_CALL_IF_KILL_VECTOR                4
#define RTNUM_CALL_IF_CONSOLE_PROCS              5
#define RTNUM_CALL_IF_DEBUG_PROCS                6
#define RTNUM_CALL_IF_FLUSH_DCACHE               7
#define RTNUM_CALL_IF_FLUSH_ICACHE               8
#define RTNUM_CALL_IF_CPU_DATA                   9
#define RTNUM_CALL_IF_BOARD_DATA                 10
#define RTNUM_CALL_IF_SYSINFO                    11
#define RTNUM_CALL_IF_SET_DEBUG_COMM             12
#define RTNUM_CALL_IF_SET_CONSOLE_COMM           13
#define RTNUM_CALL_IF_SET_SERIAL_BAUD            14
#define RTNUM_CALL_IF_DBG_SYSCALL                15
#define RTNUM_CALL_IF_RESET                      16
#define RTNUM_CALL_IF_CONSOLE_INTERRUPT_FLAG     17
#define RTNUM_CALL_IF_DELAY_US                   18
#define RTNUM_CALL_IF_DBG_DATA                   19

#define RTNUM_CALL_IF_LAST_ENTRY                 RTNUM_CALL_IF_DBG_DATA

#define RTNUM_CALL_IF_INSTALL_BPT_FN             35

#define RTNUM_CALL_IF_TABLE_SIZE                 64

externC volatile RT_ADDRWORD hal_virtual_vector_table[RTNUM_CALL_IF_TABLE_SIZE];

// Table version contains version information for both the CALL table
// itself (the number of the last active entry in the table), and the
// COMM table (the size of the table).
#define RTNUM_CALL_IF_TABLE_VERSION_CALL         RTNUM_CALL_IF_LAST_ENTRY
#define RTNUM_CALL_IF_TABLE_VERSION_CALL_HACK    (RTNUM_CALL_IF_TABLE_SIZE+1)
#define RTNUM_CALL_IF_TABLE_VERSION_CALL_MAX     RTNUM_CALL_IF_TABLE_SIZE
#define RTNUM_CALL_IF_TABLE_VERSION_COMM         RTNUM_COMM_IF_TABLE_SIZE
#define RTNUM_CALL_IF_TABLE_VERSION_CALL_MASK    0x0000ffff
#define RTNUM_CALL_IF_TABLE_VERSION_COMM_MASK    0xffff0000
#define RTNUM_CALL_IF_TABLE_VERSION_COMM_shift   16


// These are special debug/console procs IDs
// QUERY_CURRENT will cause the ID of the currently selected proc ID to be
//               returned.
// EMPTY         this is the ID used for an empty procs table (i.e, NULL
//               pointer)
// MANGLER       selects the procs space reserved for the console mangler
//               allowing the application to temporarily disable mangling
//               or temporarily switch in different console procs.
#define RTNUM_CALL_IF_SET_COMM_ID_QUERY_CURRENT -1
#define RTNUM_CALL_IF_SET_COMM_ID_EMPTY         -2
#define RTNUM_CALL_IF_SET_COMM_ID_MANGLER       -3

// The below is a (messy) attempt at adding some type safety to the
// above array. At the same time, the accessors allow the
// implementation to be easily changed in the future (both tag->data
// table and structure implementations have been suggested).

typedef int __call_if_version_t;
typedef void* __call_if_ictrl_table_t;
typedef void* __call_if_exc_table_t;
typedef bsp_handler_t *__call_if_dbg_vector_t;
typedef bsp_handler_t __call_if_kill_vector_t;
typedef hal_virtual_comm_table_t *__call_if_console_procs_t;
typedef hal_virtual_comm_table_t *__call_if_debug_procs_t;
typedef void (*__call_if_flush_dcache_t)(void *__p, int __nbytes);
typedef void (*__call_if_flush_icache_t)(void *__p, int __nbytes);
typedef void* __call_if_cpu_data_t;
typedef void* __call_if_board_data_t;
typedef int (*__call_if_sysinfo_t)(int __id, void* __ap);
//typedef int (*__call_if_sysinfo_t)(enum bsp_info_id __id, va_list __ap);
typedef int (*__call_if_set_debug_comm_t)(int __comm_id);
typedef int (*__call_if_set_console_comm_t)(int __comm_id);
typedef int (*__call_if_set_serial_baud_t)(int __comm_id, int __baud);
typedef void* __call_if_dbg_data_t;
typedef int (*__call_if_dbg_syscall_t) (enum dbg_syscall_ids id,
                                        union dbg_thread_syscall_parms  *p );
typedef void (*__call_if_reset_t)(void);
typedef int __call_if_console_interrupt_flag_t;
typedef void (*__call_if_delay_us_t)(rt_int32 usecs);
typedef void (*__call_if_install_bpt_fn_t)(void *__epc);

#ifndef RTACC_CALL_IF_DEFINED

#define RTACC_CALL_IF_VERSION() \
 ((__call_if_version_t)hal_virtual_vector_table[RTNUM_CALL_IF_VERSION])
#define RTACC_CALL_IF_VERSION_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_VERSION]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_ICTRL_TABLE() \
 ((__call_if_ictrl_table_t)hal_virtual_vector_table[RTNUM_CALL_IF_ICTRL_TABLE])
#define RTACC_CALL_IF_ICTRL_TABLE_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_ICTRL_TABLE]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_EXC_TABLE() \
 ((__call_if_exc_table_t)hal_virtual_vector_table[RTNUM_CALL_IF_EXC_TABLE])
#define RTACC_CALL_IF_EXC_TABLE_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_EXC_TABLE]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_DBG_VECTOR() \
 ((__call_if_dbg_vector_t)hal_virtual_vector_table[RTNUM_CALL_IF_DBG_VECTOR])
#define RTACC_CALL_IF_DBG_VECTOR_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_DBG_VECTOR]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_KILL_VECTOR() \
 ((__call_if_kill_vector_t)hal_virtual_vector_table[RTNUM_CALL_IF_KILL_VECTOR])
#define RTACC_CALL_IF_KILL_VECTOR_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_KILL_VECTOR]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_CONSOLE_PROCS() \
 ((__call_if_console_procs_t)hal_virtual_vector_table[RTNUM_CALL_IF_CONSOLE_PROCS])
#define RTACC_CALL_IF_CONSOLE_PROCS_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_CONSOLE_PROCS]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_DEBUG_PROCS() \
 ((__call_if_debug_procs_t)hal_virtual_vector_table[RTNUM_CALL_IF_DEBUG_PROCS])
#define RTACC_CALL_IF_DEBUG_PROCS_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_DEBUG_PROCS]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_FLUSH_DCACHE(_p_, _n_) \
 ((__call_if_flush_dcache_t)hal_virtual_vector_table[RTNUM_CALL_IF_FLUSH_DCACHE])((_p_), (_n_))
#define RTACC_CALL_IF_FLUSH_DCACHE_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_FLUSH_DCACHE]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_FLUSH_ICACHE(_p_, _n_) \
 ((__call_if_flush_icache_t)hal_virtual_vector_table[RTNUM_CALL_IF_FLUSH_ICACHE])((_p_), (_n_))
#define RTACC_CALL_IF_FLUSH_ICACHE_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_FLUSH_ICACHE]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_CPU_DATA() \
 ((__call_if_cpu_data_t)hal_virtual_vector_table[RTNUM_CALL_IF_CPU_DATA])
#define RTACC_CALL_IF_CPU_DATA_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_CPU_DATA]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_BOARD_DATA() \
 ((__call_if_board_data_t)hal_virtual_vector_table[RTNUM_CALL_IF_BOARD_DATA])
#define RTACC_CALL_IF_BOARD_DATA_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_BOARD_DATA]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_SYSINFO(_i_, _a_) \
 ((__call_if_sysinfo_t)hal_virtual_vector_table[RTNUM_CALL_IF_SYSINFO])((_i_), (_a_))
#define RTACC_CALL_IF_SYSINFO_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_SYSINFO]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_SET_DEBUG_COMM(_i_) \
 ((__call_if_set_debug_comm_t)hal_virtual_vector_table[RTNUM_CALL_IF_SET_DEBUG_COMM])((_i_))
#define RTACC_CALL_IF_SET_DEBUG_COMM_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_SET_DEBUG_COMM]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_SET_CONSOLE_COMM(_i_) \
 ((__call_if_set_console_comm_t)hal_virtual_vector_table[RTNUM_CALL_IF_SET_CONSOLE_COMM])((_i_))
#define RTACC_CALL_IF_SET_CONSOLE_COMM_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_SET_CONSOLE_COMM]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_SET_SERIAL_BAUD(_i_, _b_) \
 ((__call_if_set_serial_baud_t)hal_virtual_vector_table[RTNUM_CALL_IF_SET_SERIAL_BAUD])((_i_), (_b_))
#define RTACC_CALL_IF_SET_SERIAL_BAUD_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_SET_SERIAL_BAUD]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_DBG_DATA() \
 ((__call_if_dbg_data_t)hal_virtual_vector_table[RTNUM_CALL_IF_DBG_DATA])
#define RTACC_CALL_IF_DBG_DATA_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_DBG_DATA]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_DBG_SYSCALL() \
 ((__call_if_dbg_syscall_t)hal_virtual_vector_table[RTNUM_CALL_IF_DBG_SYSCALL])
#define RTACC_CALL_IF_DBG_SYSCALL_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_DBG_SYSCALL]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_RESET() \
 ((__call_if_reset_t)hal_virtual_vector_table[RTNUM_CALL_IF_RESET])()
#define RTACC_CALL_IF_RESET_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_RESET]=(RT_ADDRWORD)(_x_)
#define RTACC_CALL_IF_RESET_GET() \
 ((__call_if_reset_t)hal_virtual_vector_table[RTNUM_CALL_IF_RESET])

#define RTACC_CALL_IF_CONSOLE_INTERRUPT_FLAG() \
 ((__call_if_console_interrupt_flag_t)hal_virtual_vector_table[RTNUM_CALL_IF_CONSOLE_INTERRUPT_FLAG])
#define RTACC_CALL_IF_CONSOLE_INTERRUPT_FLAG_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_CONSOLE_INTERRUPT_FLAG]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_DELAY_US(_u_) \
 ((__call_if_delay_us_t)hal_virtual_vector_table[RTNUM_CALL_IF_DELAY_US])((_u_))
#define RTACC_CALL_IF_DELAY_US_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_DELAY_US]=(RT_ADDRWORD)(_x_)

#define RTACC_CALL_IF_INSTALL_BPT_FN(_e_) \
 ((__call_if_install_bpt_fn_t)hal_virtual_vector_table[RTNUM_CALL_IF_INSTALL_BPT_FN])((_e_))
#define RTACC_CALL_IF_INSTALL_BPT_FN_SET(_x_) \
 hal_virtual_vector_table[RTNUM_CALL_IF_INSTALL_BPT_FN]=(RT_ADDRWORD)(_x_)

#endif // RTACC_CALL_IF_DEFINED

//--------------------------------------------------------------------------
// Diag wrappers.
externC void hal_if_diag_init(void);
externC void hal_if_diag_write_char(char c);
externC void hal_if_diag_read_char(char *c);

//--------------------------------------------------------------------------
// Ctrl-c support.
externC rt_uint32 hal_ctrlc_isr(RT_ADDRWORD vector, RT_ADDRWORD data);
externC rt_bool   hal_ctrlc_check(RT_ADDRWORD vector, RT_ADDRWORD data);

#define HAL_CTRLC_ISR hal_ctrlc_isr
#define HAL_CTRLC_CHECK hal_ctrlc_check

#else

#if defined(RTDBG_HAL_DEBUG_GDB_BREAK_SUPPORT) \
    || defined(RTDBG_HAL_DEBUG_GDB_CTRLC_SUPPORT)
// Then other code might invoke this macro
#define HAL_CTRLC_CHECK(a1,a2) (0) // Nothing, no CTRLC here
#endif

#endif // RTSEM_HAL_VIRTUAL_VECTOR_SUPPORT

//--------------------------------------------------------------------------
// Functions provided by the HAL interface.
externC void hal_if_init(void);

//-----------------------------------------------------------------------------
#endif // RTONCE_HAL_HAL_IF_H
// End of hal_if.h
